using System;
using System.Data;
using System.Data.OleDb;
using Shop.Common;

using Shop.DAL.SalesDataSetTableAdapters;

namespace Shop.BLL
{
	public interface IProduct
	{
        DataSet GetByID(string user, string pass, int ProductID);
        DataSet FindByName(string user, string pass, string NamePattern);
        int GetStock(string user, string pass, int ProductID, out ShopStatus status);
        ShopStatus InStock(string user, string pass, int ProductID, int Quant);
        ShopStatus AddStock(string user, string pass, int ProductID, int Quant);
	}

	public class Product : IProduct
	{
		#region IProduct
		/// <summary>
		/// pesquisa um determinado produto por cdigo de produto
		/// </summary>
		/// <param name="User">username de login na aplicao</param>
		/// <param name="Pass">password de login na aplicao</param>
		/// <returns>a informao do produto</returns>
		public DataSet GetByID(string user, string pass, int ProductID)
		{
            ShopStatus status = ShopStatus.Error;
            DataSet rs = null;
			try
			{
				// validar utilizar
                User u = new User();
                status = u.Validate(user, pass);
				if (status != ShopStatus.Ok)
					return null;

                ProductsTableAdapter p = new ProductsTableAdapter();
                DAL.SalesDataSet ds = new Shop.DAL.SalesDataSet(); 
                p.FillByID(ds.Products, ProductID);
                return ds;
			}
			catch(ShopException ex)
			{
				System.Console.WriteLine("EXCEPO no mtodo <GetById> da classe 'Product': " + ex.Message);
				status = ShopStatus.DatabaseError;
                return null;
			}
		}

		/// <summary>
		/// pesquisa produtos por nome
		/// </summary>
		/// <param name="User">username de login na aplicao</param>
		/// <param name="Pass">password de login na aplicao</param>
		/// <returns>informao dos produtos que obedeam ao critrio de busca</returns>
        public DataSet FindByName(string user, string pass, string NamePattern)
		{
			DataSet rs = null;
			ShopStatus status = ShopStatus.Ok;

			try
			{	
				// validar o utilizador
                User u = new User();
                status = u.Validate(user, pass);
                if (status != ShopStatus.Ok)
                    return null;

                ProductsTableAdapter p = new ProductsTableAdapter();
                DAL.SalesDataSet ds = new Shop.DAL.SalesDataSet(); 
                p.FillByName(ds.Products, NamePattern);
                return ds;
			}
			catch(ShopException ex)
			{
				System.Console.WriteLine("EXCEPO no mtodo <FindByName> da classe 'Product': " + ex.Message);
				status = ShopStatus.DatabaseError;
                return null;
			}
		}

		/// <summary>
		/// devolve a quantidade em stock de um determinado produto
		/// </summary>
		/// <param name="User">username de login na aplicao</param>
		/// <param name="Pass">password de login na aplicao</param>
		/// <param name="ProductID">cdigo de produto a pesquizar</param>
		/// <param name="status"></param>
		/// <returns></returns>
        public int GetStock(string user, string pass, int ProductID, out ShopStatus status)
		{
			int qt;

			try
			{
                // validar o utilizador
                User u = new User();
                status = u.Validate(user, pass);
                if (status != ShopStatus.Ok)
                    return int.MinValue;

                ProductsTableAdapter p = new ProductsTableAdapter();
                DAL.SalesDataSet.ProductsDataTable ds = p.GetDataByID(ProductID);
                if (ds.Count != 1)
                {
                    status = ShopStatus.InvalidProductID;
                    return int.MinValue;
                }

				status = ShopStatus.Ok;
                return ds[0].StockQtd;
			}
            catch (ShopException ex)
			{
				System.Console.WriteLine("EXCEPO no mtodo <InStock> da classe 'Product': " + ex.Message);
				status = ShopStatus.DatabaseError;
                return int.MinValue;
			}
		}

		/// <summary>
		/// verifica se existe uma determinada quantidade de um produto em armazem 
		/// </summary>
		/// <param name="User">username de login na aplicao</param>
		/// <param name="Pass">password de login na aplicao</param>
		/// <param name="ProductID">cdigo de produto a pesquizar</param>
		/// <param name="Quant">quantidade desejada de produto</param>
		/// <returns>cdigo de estado relativo  operao</returns></returns></returns>
        public ShopStatus InStock(string user, string pass, int ProductID, int Quant)
		{
			ShopStatus status;

			try
			{
                // validar o utilizador
                User u = new User();
                status = u.Validate(user, pass);
                if (status != ShopStatus.Ok)
                    return status;

                ProductsTableAdapter p = new ProductsTableAdapter();
                DAL.SalesDataSet.ProductsDataTable ds = p.GetDataByID(ProductID);
                if (ds.Count != 1)
                    return ShopStatus.InvalidProductID;

                int qt = ds[0].StockQtd;
                if (qt >= Quant)
                    return ShopStatus.Ok;
                else
                    return ShopStatus.InsuficientStock;
			}
            catch (ShopException ex)
			{
				System.Console.WriteLine("EXCEPO no mtodo <InStock> da classe 'Product': " + ex.Message);
				return ShopStatus.DatabaseError; // ShopStatus.NotOk;
			}
		}

        public ShopStatus AddStock(string user, string pass, int ProductID, int Quant)
		{
			ShopStatus status;

			// validar parametros de entrada
			if (Quant <= 0)
				return ShopStatus.InvalidArgument;

			try
			{
                // validar o utilizador
                User u = new User();
                status = u.Validate(user, pass);
                if (status != ShopStatus.Ok)
                    return status;

                ProductsTableAdapter p = new ProductsTableAdapter();
                if (p.UpdateStock(ProductID, Quant) != 1)
                    status = ShopStatus.InvalidProductID;
			}
            catch (ShopException ex)
			{
				System.Console.WriteLine("EXCEPO no mtodo <AddStock> class 'Product': " + ex.Message);
				status = ShopStatus.DatabaseError;
			}

			return status;
		}

		public DataSet GetCatalog(string customerUserName, string customerPassword, out ShopStatus status)
		{
			try
			{
                status = ShopStatus.InvalidLogin;
				// validar o utilizador
				int customerID;
                Customer u = new Customer();
				if (u.ValidateCustomerLogin(customerUserName, customerPassword) != ShopStatus.Ok)
					return null;

                status = ShopStatus.InvalidProductID;
                ProductsTableAdapter p = new ProductsTableAdapter();
                DAL.SalesDataSet ds = new Shop.DAL.SalesDataSet();
                p.Fill(ds.Products);
				ds.DataSetName = "Catalog";
                status = ShopStatus.Ok;
                return ds;
			}
            catch (ShopException ex)
			{
				System.Console.WriteLine("EXCEPO no mtodo <GetCatalog> da classe 'Product': " + ex.Message);
				status = ShopStatus.DatabaseError;
                return null;
			}
        }
        #endregion
	}
}
