using System;
using System.Data;
using System.Data.OleDb;
using Shop.Common;

namespace Shop.BLL
{
	public interface ISale
	{
		DataSet CreateDetails(string user, string pass);
		ShopStatus Add(string user, string pass,int CustomerID, DateTime Date, DataSet RsDetails);
		DataSet GetByCustomer(string user, string pass,int CustomerID);
		DataSet GetAll(string user, string pass);
		DataSet GetById(string user, string pass,int saleid);
		ShopStatus GenerateWordDocumentInvoice(string user, string pass,int saleid, out string filename);
	}

	/// <summary>
	/// Summary description for Sale.
	/// </summary>
	public class Sale : ISale
	{
		#region ISale

		/// <summary>
		/// pesquisa clientes por nome
		/// </summary>
		/// <param name="User">username de login na aplicao</param>
		/// <param name="Pass">password de login na aplicao</param>
		/// <returns>um dataset vazio para a palicao cliente preencher</returns>
		public DataSet CreateDetails(string user, string pass)
		{
            return UtilDB.GetByID("SaleDetails", "SaleId", -1);
		}

		/// <summary>
		/// adiciona uma venda ao sistema
		/// </summary>
		/// <param name="User">username de login na aplicao</param>
		/// <param name="Pass">password de login na aplicao</param>
		/// <param name="RsDetails">um dataset com as linhas de detalhe da venda</param>
		/// <returns>o cdigo de status relativo  operao</returns>
		public ShopStatus Add(string user, string pass,int CustomerID, DateTime Date, DataSet RsDetails)
		{
			ShopStatus ret = ShopStatus.NotOk;
            OleDbConnection conn = null;
            OleDbTransaction tx = null;

			// validaes
			if (user.Length == 0 || pass.Length == 0)
				return ShopStatus.InvalidArgument;
			if (RsDetails == null || RsDetails.Tables.Count != 1 || RsDetails.Tables[0].TableName != "SaleDetails")
				return ShopStatus.InvalidArgument;
			if (RsDetails.Tables["SaleDetails"].Rows.Count == 0)
				return ShopStatus.InvalidArgument;

			try
			{
				// criar objecto de conexo  base de dados
				conn = new OleDbConnection(UtilDB.CONN);
				conn.Open();

				// iniciar transaco
				tx = conn.BeginTransaction();

				// validar o utilizador
                if (UtilDB.ValidateUser(conn, tx, user, pass) != ShopStatus.Ok)
					return ShopStatus.InvalidLogin;

				// validar cliente
                DataSet dsCust = UtilDB.GetByID(conn, tx, "Customers", "CustomerID", CustomerID);
				if (dsCust.Tables["Customers"].Rows.Count == 0)
					return ShopStatus.InvalidCustomerID;

				// validar os produtos e quantidades pedidas
//				if (RsDetails.Tables["SaleDetails"].Rows.Count == 0)
//					return ShopStatus.InvalidArgument;

				int[] listaProd = new int[RsDetails.Tables["SaleDetails"].Rows.Count];
				int i = 0;
				foreach (DataRow regDetails in RsDetails.Tables["SaleDetails"].Rows)
				{
					if (regDetails.IsNull("ProductID"))
						return ShopStatus.InvalidArgument; //INVALID_PRODUCT_ID;
					if (regDetails.IsNull("Quantity"))
						return ShopStatus.InvalidArgument;

					// verificar se este produto tem quantidade suficiente em stock
					// e se existe!
					ret = InStock(conn, tx, (int)regDetails["ProductID"], (int)regDetails["Quantity"]);
					if (ret != ShopStatus.Ok)
						return ret;

					// guardar o cdigo de produto para pesquisa
					listaProd[i++] = (int)regDetails["ProductID"];
				}

				// criar um DataSet de produtos e preencher apenas com os produtos desejados
                DataSet pDS = UtilDB.GetByID(conn, tx, "Products", "ProductID", listaProd);
				if (pDS == null || pDS.Tables[0].Rows.Count != RsDetails.Tables["SaleDetails"].Rows.Count)
					return ShopStatus.InvalidProductID;

				// calcular custo por linha
				foreach (DataRow regDetails in RsDetails.Tables["SaleDetails"].Rows)
				{
					// procurar o produto desejado
					DataRow[] row = pDS.Tables["Products"].Select("ProductId=" + regDetails["ProductID"].ToString());

					// calcular custo desta linha da venda
					regDetails["TotalCost"] = (int)regDetails["Quantity"] * (float)row[0]["UnitPrice"];

					/**
					//
					// actualizar stock no DataSet para posterior update  BD
					//
					row[0]["StockQtd"] = (int)row[0]["StockQtd"] - (int)regDetails["Quantity"];
					**/
				}


				// ======================================
				//
				// tratar das inseres
				//
				// ======================================


				// criar comando para insero
				OleDbCommand cmd = new OleDbCommand("INSERT INTO sales (customerid, dateofsale) VALUES (?, ?)", conn);
				cmd.Transaction = tx;
				cmd.Parameters.Add("custid", CustomerID);
				OleDbParameter parm = cmd.Parameters.Add("dt", OleDbType.Date);
				parm.Value = Date;

				// inserir registo 
				cmd.ExecuteNonQuery();

				// obter id autogerado
				OleDbCommand cmd2 = new OleDbCommand("SELECT @@IDENTITY", conn);
				cmd2.Transaction = tx;
				int id = (int)cmd2.ExecuteScalar();

				// inserir chave estrangeira nas linhas de detalhe
				foreach (DataRow regDetails in RsDetails.Tables["SaleDetails"].Rows)
					regDetails["SaleID"] = id;

				// inserir linhas de detalhe
				OleDbDataAdapter da = new OleDbDataAdapter("SELECT * FROM SaleDetails", conn);
				da.SelectCommand.Transaction = tx;
				OleDbCommandBuilder cb = new OleDbCommandBuilder(da);
				da.Update(RsDetails, "SaleDetails");

				// actualizar quantidades de produto em stock
				OleDbCommand cmd3 = new OleDbCommand("UPDATE Products SET StockQtd=StockQtd-? WHERE ProductId=?", conn);
				cmd3.Transaction = tx;
				OleDbParameter p1 = cmd3.Parameters.Add("qtd", OleDbType.Integer);
				OleDbParameter p2 = cmd3.Parameters.Add("pid", OleDbType.Integer);
				foreach (DataRow regDetails in RsDetails.Tables["SaleDetails"].Rows)
				{
					p1.Value = regDetails["Quantity"];
					p2.Value = regDetails["ProductID"];

					cmd3.ExecuteNonQuery();
				}
                tx.Commit();
			}
			catch(OleDbException ex)
			{
                if (tx != null)
                    tx.Rollback();

				System.Console.WriteLine("EXCEPO no mtodo <Add> da classe 'Sale': " + ex.Message);
				//status = ShopStatus.DatabaseError;
			}
			finally
			{
                if (conn.State != ConnectionState.Closed)
				    conn.Close();   
			}

			return ret;
		}

		public DataSet GetByCustomer(string user, string pass,int CustomerID)
		{
			DataSet ds = null;
			OleDbConnection conn = null;

			try
			{				
				//
				// criar objecto DataSet
				//
				ds = new DataSet();

				//
				// criar objecto de conexo  base de dados e abrir a conexo
				//
				conn = new OleDbConnection(UtilDB.CONN);
				conn.Open();

				//
				// validar o utilizador
				//
                if (UtilDB.ValidateUser(conn, null, user, pass) != ShopStatus.Ok)
					return null;

				//
				// criar um Datadapter para executar o comando e devolver o dataset
				// cabealhos das vendas
				//
				string sSqlCmd = "Select * From vwTotalBySale Where CustomerID = ?";
				OleDbDataAdapter oAdapter = new OleDbDataAdapter(sSqlCmd, conn);
				oAdapter.SelectCommand.Parameters.Add("custid", CustomerID);

				//
				// executar o comando e preencher um Dataset
				//
				oAdapter.Fill(ds, "Sales");

				//
				// criar um Datadapter para executar o comando e devolver o dataset
				// linhas de detalhe
				//
				sSqlCmd = "SELECT * FROM vwExpandedSaleDetails WHERE SaleId IN (SELECT SaleID FROM Sales WHERE CustomerId=?)";
				oAdapter = new OleDbDataAdapter(sSqlCmd, conn);
				oAdapter.SelectCommand.Parameters.Add("custid", CustomerID);

				//
				// executar o comando e preencher um Dataset
				//
				oAdapter.Fill(ds, "SaleDetails");

				//
				// criar relao entre tabelas
				//
				ds.Relations.Add("Details", ds.Tables["Sales"].Columns["SaleId"], ds.Tables["SaleDetails"].Columns["SaleId"]);
			}
			catch(OleDbException ex)
			{
				//
				// tratar a excepo!!!!
				//
				System.Console.WriteLine("EXCEPO no mtodo 'GetByCustomer': " + ex.Message);
				//status = ShopStatus.DatabaseError;
				ds = null;
			}
			finally
			{
				//
				// fechar a conexo
				//
				if (conn.State == ConnectionState.Open)
					conn.Close();
			}

			return ds;

		}

		public DataSet GetAll(string user, string pass)
		{
			DataSet ds = null;
			OleDbConnection conn = null;

			try
			{				
				//
				// criar objecto DataSet
				//
				ds = new DataSet();

				//
				// criar objecto de conexo  base de dados e abrir a conexo
				//
				conn = new OleDbConnection(UtilDB.CONN);
				conn.Open();

				//
				// validar o utilizador
				//
                if (UtilDB.ValidateUser(conn, null, user, pass) != ShopStatus.Ok)
					return null;

				//
				// criar um Datadapter para executar o comando e devolver o dataset
				// cabealhos das vendas
				//
				string sSqlCmd = "Select * From vwTotalBySale";
				OleDbDataAdapter oAdapter = new OleDbDataAdapter(sSqlCmd, conn);

				//
				// executar o comando e preencher um Dataset
				//
				oAdapter.Fill(ds, "Sales");

				//
				// criar um Datadapter para executar o comando e devolver o dataset
				// linhas de detalhe
				//
				sSqlCmd = "Select * From vwExpandedSaleDetails";
				oAdapter = new OleDbDataAdapter(sSqlCmd, conn);

				//
				// executar o comando e preencher um Dataset
				//
				oAdapter.Fill(ds, "SaleDetails");

				//
				// criar relao entre tabelas
				//
				ds.Relations.Add("Details", ds.Tables["Sales"].Columns["SaleId"], ds.Tables["SaleDetails"].Columns["SaleId"]);
			}
			catch(OleDbException ex)
			{
				//
				// tratar a excepo!!!!
				//
				System.Console.WriteLine("EXCEPO no mtodo 'GetAll': " + ex.Message);
				//status = ShopStatus.DatabaseError;
				ds = null;
			}
			finally
			{
				//
				// fechar a conexo
				//
				if (conn.State == ConnectionState.Open)
					conn.Close();
			}

			return ds;

		}

		#endregion

		#region ISaleOffice Members

		private const string OUTDIR_DOC = @"c:\temp\WinShop\";

		public DataSet GetById(string user, string pass, int saleid)
		{
			DataSet ds = null;
			OleDbConnection conn = null;

			try
			{				
				//
				// criar objecto DataSet
				//
				ds = new DataSet();

				//
				// criar objecto de conexo  base de dados e abrir a conexo
				//
				conn = new OleDbConnection(UtilDB.CONN);
				conn.Open();

				//
				// validar o utilizador
				//
                if (UtilDB.ValidateUser(conn, null, user, pass) != ShopStatus.Ok)
					return null;

				//
				// criar um Datadapter para executar o comando e devolver o dataset
				// cabealhos das vendas
				//
				string sSqlCmd = "SELECT * FROM vwTotalBySale WHERE SaleId=?";
				OleDbDataAdapter oAdapter = new OleDbDataAdapter(sSqlCmd, conn);
				oAdapter.SelectCommand.Parameters.Add("sid", saleid);

				//
				// executar o comando e preencher um Dataset
				//
				oAdapter.Fill(ds, "Sales");

				//
				// criar um Datadapter para executar o comando e devolver o dataset
				// linhas de detalhe
				//
				sSqlCmd = "SELECT * FROM vwExpandedSaleDetails WHERE SaleId=?";
				oAdapter = new OleDbDataAdapter(sSqlCmd, conn);
				oAdapter.SelectCommand.Parameters.Add("sid", saleid);

				//
				// executar o comando e preencher um Dataset
				//
				oAdapter.Fill(ds, "SaleDetails");

				//
				// criar relao entre tabelas
				//
				ds.Relations.Add("Details", ds.Tables["Sales"].Columns["SaleId"], ds.Tables["SaleDetails"].Columns["SaleId"]);
			}
			catch(OleDbException ex)
			{
				//
				// tratar a excepo!!!!
				//
				System.Console.WriteLine("EXCEPO no mtodo 'GetSaleById': " + ex.Message);
				//status = ShopStatus.DatabaseError;
				ds = null;
			}
			finally
			{
				//
				// fechar a conexo
				//
				if (conn.State == ConnectionState.Open)
					conn.Close();
			}

			return ds;
		}

		public ShopStatus GenerateWordDocumentInvoice(string User, string Pass, int saleid, out string filename)
		{
			object missingValue = Type.Missing; 
			object uchar = Word.WdUnits.wdCharacter;
			object uline = Word.WdUnits.wdLine;
			object ucol = Word.WdUnits.wdColumn;
			object i1 = 1;
			object i4 = 4;
			object bfalse = false;
			object btrue = true;

			filename = null;

			// obter factura
			DataSet dsFactura = GetById(User, Pass, saleid);
			if (dsFactura == null)
				return ShopStatus.Error;

			int nFactura = saleid; //devia existir um campo separado para o numero de facturo que tem que ser sequencial e sem "buracos";
			DateTime dataFactura = (DateTime)dsFactura.Tables["Sales"].Rows[0]["DateOfSale"];
			string customerName = (string)dsFactura.Tables["Sales"].Rows[0]["Name"];
			double totalFactura = (double)dsFactura.Tables["Sales"].Rows[0]["SaleTotal"];

			// gerar nome de ficheiro
			object _in_filename = OUTDIR_DOC + "Factura-" + nFactura.ToString() + ".doc";

			// criar objecto Word
			Word.Application w = new Word.ApplicationClass();

			// criar novo documento
			w.Documents.Add(ref missingValue, ref missingValue, ref missingValue, ref missingValue);

			//w.Visible = true;	// s para teste/"debug"!!!!!

			w.Selection.TypeText("Factura n ");
			w.Selection.TypeText(nFactura.ToString());
			w.Selection.TypeParagraph();
			w.Selection.TypeText("Data : ");
			w.Selection.TypeText(dataFactura.ToLongDateString());
			w.Selection.TypeParagraph();
			w.Selection.TypeText("Cliente : ");
			w.Selection.TypeText(customerName);
			w.Selection.TypeParagraph();
			w.Selection.TypeParagraph();

			//
			// criar tabela (cabealho)
			//
			object tb = Word.WdDefaultTableBehavior.wdWord9TableBehavior;
			object afb = Word.WdAutoFitBehavior.wdAutoFitWindow;
			w.ActiveDocument.Tables.Add(w.Selection.Range, 1, 4, ref tb, ref afb);
			
			w.Selection.TypeText("Quantidade");
			w.Selection.MoveRight(ref uchar, ref i1, ref missingValue);
			w.Selection.TypeText("Designao");
			w.Selection.MoveRight(ref uchar, ref i1, ref missingValue);
			w.Selection.TypeText("P.U:");
			w.Selection.MoveRight(ref uchar, ref i1, ref missingValue);
			w.Selection.TypeText("sub-total");

			//
			// adicionar as linhas de factura
			//
			foreach (DataRow dr in dsFactura.Tables["SaleDetails"].Rows)
			{
				// adicionar linha  tabela
				w.ActiveDocument.Tables[1].Rows.Add(ref missingValue);
				w.Selection.MoveDown(ref uline, ref i1, ref missingValue);
				w.Selection.MoveLeft(ref uchar, ref i4, ref missingValue);

				w.Selection.TypeText(dr["Quantity"].ToString());
				w.Selection.MoveRight(ref uchar, ref i1, ref missingValue);
				w.Selection.TypeText(dr["Description"].ToString());
				w.Selection.MoveRight(ref uchar, ref i1, ref missingValue);
				w.Selection.TypeText(dr["UnitPrice"].ToString());
				w.Selection.MoveRight(ref uchar, ref i1, ref missingValue);
				w.Selection.TypeText(dr["TotalCost"].ToString());
			}
			w.Selection.MoveDown(ref uline, ref i1, ref missingValue);

			w.Selection.TypeParagraph();
			w.Selection.TypeText("Total :");
			w.Selection.TypeText(totalFactura.ToString());
			w.Selection.TypeParagraph();

			//
			// formatar tabela
			//
			object tf = Word.WdTableFormat.wdTableFormatProfessional;
			w.ActiveDocument.Tables[1].AutoFormat(ref tf, ref btrue, ref btrue, ref btrue, ref btrue, ref btrue, ref bfalse, ref bfalse, ref bfalse, ref bfalse);

			// rodap da factura
			w.Selection.TypeParagraph();
			w.Selection.TypeText("Generated by WinShop 1.1");

			// gravar ficheiro
			w.ActiveDocument.SaveAs(ref _in_filename, ref missingValue, ref missingValue, ref missingValue, ref missingValue, ref missingValue, ref missingValue, ref missingValue, ref missingValue, ref missingValue, ref missingValue, ref missingValue, ref missingValue, ref missingValue, ref missingValue, ref missingValue);

			//terminar Word
			w.Quit(ref missingValue, ref missingValue, ref missingValue);


			filename = (string)_in_filename;
			return ShopStatus.Ok;
		}

		#endregion

        /// <summary>
        /// verifica se existe uma determinada quantidade de um produto em armazem.
        /// este mtodo no faz parte do interface pblico da classe, apenas  
        /// utilizado internamente na ClassLibrary
        /// </summary>
        /// <param name="ProductID">cdigo de produto a pesquizar</param>
        /// <param name="Quant">quantidade desejada de produto</param>
        /// <returns>cdigo de estado relativo  operao</returns></returns></returns>
        private ShopStatus InStock(OleDbConnection conn, OleDbTransaction tx, int ProductID, int Quant)
        {
            ShopStatus status = ShopStatus.Ok;

            //
            // efectuar o select
            //
            DataSet rs = UtilDB.GetByID(conn, tx, "Products", "ProductID", ProductID);
            if (rs == null)
                return ShopStatus.DatabaseError;

            //
            // validar stock
            //
            if (rs.Tables["products"].Rows.Count > 0)
                status = ((int)(rs.Tables["Products"].Rows[0]["StockQtd"]) >= Quant ? ShopStatus.Ok : ShopStatus.InsuficientStock);
            else
                status = ShopStatus.InvalidProductID;

            return status;
        }

        private class UtilDB
        {
            public static string CONN
            {
                get
                {
                    //return DEFAULT_CONN + DEFAULT_DBPATH; 
                    return System.Configuration.ConfigurationSettings.AppSettings["BD"];
                }
            }

            public static bool MayBeSqlInjection(string sqlFragment)
            {
                if (sqlFragment.IndexOf(';') > 0)
                    return true;

                return false;
            }

            #region GetByID

            static public DataSet GetByID(string table, string keyName, long key)
            {
                OleDbConnection cnx = new OleDbConnection(UtilDB.CONN);
                DataSet ds = GetByID(cnx, null, table, keyName, key);
                cnx.Close();
                return ds;
            }

            /// <summary>
            /// Obtem um dataset resultante de uma pesquisa por chave
            /// </summary>
            /// <param name="conn">a conexo a utilizar</param>
            /// <param name="table">tabela a pesquisar</param>
            /// <param name="keyName">campo chave a usar</param>
            /// <param name="key">valor da chave a pesquisar</param>
            /// <returns>um dataset do tipo pedido</returns>
            static public DataSet GetByID(OleDbConnection conn, OleDbTransaction tx, string table, string keyName, long key)
            {
                try
                {
                    DataSet ds = new DataSet();
                    string sSqlCmd = "Select * From " + table + " Where " + keyName + " = " + key.ToString();
                    OleDbDataAdapter oAdapter = new OleDbDataAdapter(sSqlCmd, conn);
                    oAdapter.SelectCommand.Transaction = tx;
                    oAdapter.Fill(ds, table);
                    return ds;
                }
                catch (OleDbException ex)
                {
                    System.Console.WriteLine("EXCEPO no mtodo 'UtilDB.GetByID': " + ex.Message);
                    throw new ShopException("UtilDB.GetById", ex);
                }
            }

            /// <summary>
            /// Obtem um dataset resultante de uma pesquisa por chave
            /// </summary>
            /// <param name="table">tabela a pesquisar</param>
            /// <param name="field">campo chave a usar</param>
            /// <param name="key">valor da chave a pesquisar</param>
            /// <returns>um dataset do tipo pedido</returns>
            static public DataSet GetByID(OleDbConnection conn, OleDbTransaction tx, string table, string keyName, int[] keys)
            {
                try
                {
                    DataSet ds = new DataSet();
                    System.Text.StringBuilder sb = new System.Text.StringBuilder("Select * From " + table + " Where " + keyName + " IN (");
                    foreach (int k in keys)
                    {
                        sb.Append(k.ToString());
                        sb.Append(",");
                    }
                    sb.Length = sb.Length - 1;
                    sb.Append(")");
                    string sSqlCmd = sb.ToString();

                    OleDbDataAdapter oAdapter = new OleDbDataAdapter(sSqlCmd, conn);
                    oAdapter.SelectCommand.Transaction = tx;
                    oAdapter.Fill(ds, table);
                    return ds;
                }
                catch (OleDbException ex)
                {
                    System.Console.WriteLine("EXCEPO no mtodo 'UtilDB.GetByID': " + ex.Message);
                    throw new ShopException("UtilDB.GetById", ex);
                }
            }

            #endregion

            #region FindByStringField

            /// <summary>
            /// Obtem um dataset resultante de uma pesquisa por chave
            /// </summary>
            /// <param name="table">tabela a pesquisar</param>
            /// <param name="field">campo chave a usar</param>
            /// <param name="key">valor da chave a pesquisar</param>
            /// <returns>um dataset do tipo pedido</returns>
            static public DataSet FindByStringField(OleDbConnection conn, string table, string field, string key)
            {

                try
                {
                    DataSet ret = new DataSet();
                    OleDbDataAdapter da = new OleDbDataAdapter("SELECT * FROM " + table + " WHERE " + field + " LIKE '" + key + "%'", conn);
                    da.Fill(ret, table);
                    return ret;
                }
                catch (OleDbException ex)
                {
                    System.Console.WriteLine("EXCEPO no componente 'UtilDB': " + ex.Message);
                    throw new ShopException("UtilDB.FindByStringField", ex);
                }
            }

            #endregion

            /// <summary>
            /// Valida um login/password no sistema
            /// </summary>
            /// <param name="User">username de login na aplicao</param>
            /// <param name="Pass">password de login na aplicao</param>
            /// <param name="isCustomer">flag indicando se deve tentar validar este utilizador como cliente</param>
            /// <returns></returns>
            static internal ShopStatus ValidateUser(OleDbConnection conn, OleDbTransaction tx, string user, string pass)
            {
                ShopStatus status = ShopStatus.Ok;
                OleDbCommand cmd = null;

                try
                {
                    cmd = new OleDbCommand("Select * from Users Where User=? and Pass=?", conn);
                    cmd.Transaction = tx;
                    cmd.Parameters.Add("@u", user);
                    cmd.Parameters.Add("@p", pass);


                    // poderia ser feito usando um count(*) e um ExecuteScalar()
                    OleDbDataReader dr = cmd.ExecuteReader();

                    // ( necessrio manter a conexo aberta pois estamos a usar um DataReader)
                    if (dr.HasRows)
                        status = ShopStatus.Ok;
                    else
                        status = ShopStatus.InvalidLogin;

                    dr.Close();
                }
                catch (OleDbException ex)
                {
                    System.Console.WriteLine("EXCEPO na validao do 'User': " + ex.Message);
                    status = ShopStatus.DatabaseError;
                }

                return status;
            }

        }
	}
}
